/* ******************************************************************************** *
 * Copyright (c) 2002 - 2004 Bright Side Factory. All rights reserved.              *
 *                                                                                  *
 * Redistribution and use in source and binary forms, with or without modification, *
 * are permitted provided that the following conditions are met:                    *
 *                                                                                  *
 * 1. Redistributions of source code must retain the above copyright notice, this   *
 * list of conditions and the following disclaimer.                                 *
 *                                                                                  *
 * 2. Redistributions in binary form must reproduce the above copyright notice,     *
 * this list of conditions and the following disclaimer in the documentation and/or *
 * other materials provided with the distribution.                                  *
 *                                                                                  *
 * 3. The end-user documentation included with the redistribution, if any, must     *
 * include the following acknowledgment: "This product includes software developed  *
 * by the Bright Side Factory (http://www.bs-factory.org/)." Alternately, this      *
 * acknowledgment may appear in the software itself, if and wherever such           *
 * third-party acknowledgments normally appear.                                     *
 *                                                                                  *
 * 4. The names "Bright Side", "BS Framework" and "Bright Side Factory" must not be *
 * used to endorse or promote products derived from this software without prior     *
 * written permission. For written permission, please contact info@bs-factory.org.  *
 *                                                                                  *
 * 5. Products derived from this software may not be called "Bright Side", nor may  *
 * "Bright Side" appear in their name, without prior written permission of the      *
 * Bright Side Factory.                                                             *
 *                                                                                  *
 * THIS SOFTWARE IS PROVIDED ''AS IS'' BY THE COPYRIGHT OWNER AND ANY EXPRESSED OR  *
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF     *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT *
 * SHALL THE COPYRIGHT OWNER OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,          *
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,   *
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,    *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF  *
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE  *
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS OFTWARE, EVEN IF ADVISED *
 * OF THE POSSIBILITY OF SUCH DAMAGE.                                               *
 *                                                                                  *
 * ================================================================================ *
 *                                                                                  *
 * This software consists of voluntary contributions made by many individuals on    *
 * behalf of the Bright Side Factory. For more information on the Bright Side       *
 * Factory, please see http://www.bs-factory.org.                                   *
 *                                                                                  *
 * ******************************************************************************** */
package org.bsf.remoteIterator.client;

import org.bsf.remoteIterator.common.RemoteIteratorResultTable;
import org.bsf.remoteIterator.server.RemoteIterator;

import javax.ejb.RemoveException;
import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.List;

/**
 * Wrapper for a remoteIterator. To be used by the clients.
 *
 * @see org.bsf.remoteIterator.server.RemoteIteratorBean
 * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
 */
public class RemoteIteratorClient implements Serializable {
    // *************************************************************************
    // ****************************** Attributes *******************************
    // *************************************************************************

    private boolean _isLast = false;
    private boolean _isFirst = true;

    /**
     * Total number of rows, -1 if not computed.
     */
    private int _rowCount = -1;

    private RemoteIterator _remote = null;

    // *************************************************************************
    // ****************************** Constructor ******************************
    // *************************************************************************

    public RemoteIteratorClient( RemoteIterator p_remoteIteratorProxy ) throws IllegalArgumentException {
        _remote = p_remoteIteratorProxy;
    }

    // *************************************************************************
    // ****************************** Absolute *********************************
    // *************************************************************************

    public synchronized Iterator absolute( int p_position ) throws RemoteException {
        if ( _remote == null ) return null;
        return getResult( _remote.absolute( p_position ) ).iterator();
    }

    public synchronized Iterator absolute( int p_position, int p_nbRecords ) throws RemoteException {
        if ( _remote == null ) return null;
        return getResult( _remote.absolute( p_position, p_nbRecords ) ).iterator();
    }
    // *************************************************************************
    // ******************************** Next ***********************************
    // *************************************************************************

    public synchronized Iterator next() throws RemoteException {
        if ( _remote == null ) return null;
        return getResult( _remote.next() ).iterator();
    }

    public synchronized Iterator next( int p_nbRecords ) throws RemoteException {
        if ( _remote == null ) return null;
        return getResult( _remote.next( p_nbRecords ) ).iterator();
    }

    // *************************************************************************
    // ****************************** Previous *********************************
    // *************************************************************************

    public synchronized Iterator previous() throws RemoteException {
        if ( _remote == null ) return null;
        return getResult( _remote.previous() ).iterator();
    }

    public synchronized Iterator previous( int p_nbRecords ) throws RemoteException {
        if ( _remote == null ) return null;
        return getResult( _remote.previous( p_nbRecords ) ).iterator();
    }

    // *************************************************************************
    // *************************** Utility Methods *****************************
    // *************************************************************************

    private synchronized List getResult( RemoteIteratorResultTable p_riResult ) {
        _isLast = p_riResult.isLast();
        _isFirst = p_riResult.isFirst();

        return p_riResult.getRows();
    }

    /**
     * Removes a RemoteItarator (releases the ressources).
     *
     * @see #hasRemoteReference
     */
    public synchronized void remove() throws RemoteException {
        try {
            if ( _remote != null ) {
                // We remove the RI
                _remote.remove();
            }
        } catch( RemoveException e ) {
            throw new RemoteException( e.getLocalizedMessage() );
        } finally {
            // No need to keep the ref to the remote
            _remote = null;
        }
    }

    /**
     * @return true if this client has a remote reference, false otherwise.
     * In other words, it returns true if the RemoteIterator has not been
     * removed, false otherwise.
     *
     * @see #remove
     */
    public synchronized boolean hasRemoteReference() {
        if ( _remote != null ) {
            return true;
        }

        return false;
    }

    // *************************************************************************
    // ******************************* Accessors *******************************
    // *************************************************************************

    /**
     * @return a List of ColumnMetaData correpsonding to the executed query.
     */
    public synchronized List getColumnMetatData() throws RemoteException {
        if ( _remote == null ) {
            return null;
        }

        return _remote.getColumnMetaData();
    }

    /**
     * @return the number of records returned by the request. -1 if no remote
     * is set...
     */
    public synchronized int getRecordCount() throws RemoteException {
        if ( _rowCount == -1 && _remote != null ) {
            // We update the row count (the real one for sure)
            _rowCount = _remote.getRowCount().intValue();
        }

        return _rowCount;
    }

    public synchronized boolean isLast() {
        return _isLast;
    }

    public synchronized boolean isFirst() {
        return _isFirst;
    }
}

