package org.bsf.remoteIterator.client.tableModel;

import org.bsf.remoteIterator.client.RemoteIteratorClient;

import java.rmi.RemoteException;
import java.util.Iterator;
import java.util.Vector;

/**
 * A thread to load the data in the background, using a RemoteIteratorClient.
 *
 * @see DefaultRemoteIteratorTableModel
 */
public class BackgroundLoadingThread extends Thread {
    private static Vector _instantiatedThreads = new Vector( 20, 10 );

    private RemoteIteratorClient _remoteIteratorClient = null;
    private DefaultRemoteIteratorTableModel _defaultRemoteIteratorTableModel = null;

    private boolean _shouldStop = false;
    private boolean _isFirstCall = true;

    // *************************************************************************
    // ****************************** Constructors *****************************
    // *************************************************************************

    /**
     * Instantiates a Thread that will, using the provided RemoteIteratorClient,
     * retrieve the data and insert it in the given DefaultRemoteIteratorTableModel.
     *
     * @param p_remoteIteratorClient The RIClient to use to retrieve the data.
     * @param p_tableModel The DefaultRemoteIteratorTableModel in which the data
     * will be inserted as they arrived.
     */
    public BackgroundLoadingThread( RemoteIteratorClient p_remoteIteratorClient,
                                    DefaultRemoteIteratorTableModel p_tableModel ) {
        super();

        _remoteIteratorClient = p_remoteIteratorClient;
        _defaultRemoteIteratorTableModel = p_tableModel;

        // We add it to the map of instantiated threads... For later cleanup...
        _instantiatedThreads.add( this );
    }

    // *************************************************************************
    // ********************** Overriden Thread Methods *************************
    // *************************************************************************

    public void run() {
        while ( !_shouldStop ) {
            // We retrieve the block size for this call (a different one for the
            // first call to have a quick display of data...)
            int blockSize = _defaultRemoteIteratorTableModel.getDefaultBlockSize();

            if ( _isFirstCall ) {
                blockSize = _defaultRemoteIteratorTableModel.getInitialBlockSize();

                // Not the first call anymore
                _isFirstCall = false;
            }

            try {
                // Retrieving the rows
                Iterator iterator = _remoteIteratorClient.next( blockSize );

                // Inserting them in the TableModel
                if ( _defaultRemoteIteratorTableModel != null ) {
                    _defaultRemoteIteratorTableModel.addRows( iterator );
                }

                // Do we have to go on ? (ie. Have we reached the last row ?)
                if ( _remoteIteratorClient.isLast() ) {
                    _defaultRemoteIteratorTableModel.stopLoadingData();
                } else {
                    try {
                        // We sleep the given time (if > 0)
                        int timeToSleepInMs = _defaultRemoteIteratorTableModel.getDelayBetweenRetrieval();

                        if ( timeToSleepInMs > 0 ) {
                            Thread.sleep( timeToSleepInMs );
                        }
                    } catch( InterruptedException e ) {
                        // Too bad... Should not happen though... Nothing to do...
                    }
                }
            } catch( RemoteException e ) {
                // Nothing can be done... Better luck next time ?
            }
        }

        // We remove the RI since all the data have been loaded or the loading was stopped
        removeRemoteIterator();
    }

    // *************************************************************************
    // **************************** Action Methods *****************************
    // *************************************************************************

    public void stopLoading() {
        // We tell the Thread that this is the end (my friend...)
        _shouldStop = true;

        // And we don't want to go on notify the model....
        _defaultRemoteIteratorTableModel = null;
    }

    public synchronized void removeRemoteIterator() {
        // Should have been called before (except when quitting the application
        // and that this thread is sleeping...
        stopLoading();

        try {
            if ( _remoteIteratorClient != null ) {
                // We remove the RI and we clear the ref to the client
                _remoteIteratorClient.remove();
                _remoteIteratorClient = null;

                // This thread has no more active RI
                _instantiatedThreads.remove( this );
            }
        } catch( RemoteException e ) {
            // We did what we could...
        }
    }

    /**
     *  This method is called by the model when the static cleanUp method is called
     * to prevent phantom RIs (application exited with sleeping threads...).
     */
    public static void removeAllRemainingRemoteIterator() {
        Iterator iterator = _instantiatedThreads.iterator();

        while ( iterator.hasNext() ) {
            try {
                ( (BackgroundLoadingThread) iterator.next() ).removeRemoteIterator();
            } catch( Exception e ) {
                // We don't care !!! Clean up time...
            }
        }
    }
}
