package org.bsf.remoteIterator.client.tableModel;

import org.bsf.remoteIterator.client.RemoteIteratorClient;

import javax.swing.table.AbstractTableModel;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

/**
 * A default implementation of a RemoteIteratorTableModel. You can look at it
 * (or extend it) to build your own TableModel. A smart move is to add a static
 * method that, when called, removes all the RI that are still active when called.
 */
public class DefaultRemoteIteratorTableModel extends AbstractTableModel implements RemoteIteratorTableModel {
    private int _initialBlockSize = 25;
    private int _defaultBlockSize = 250;
    private int _delayBetweenRetrieval = 0;

    private BackgroundLoadingThread _loadingThread = null;

    /**
     * Holds the rows returned by the RI.
     */
    private Vector _rows = null;

    // *************************************************************************
    // ****************************** Constructors *****************************
    // *************************************************************************

    /**
     * Creates a DefaultRemoteIteratorTableModel with a default estimated size of
     * 1000 rows. Anyway, the underlying container will be trimmed to size when
     * all the rows will have been loaded.
     *
     * @see #DefaultRemoteIteratorTableModel( int p_estimatedSize )
     */
    public DefaultRemoteIteratorTableModel() {
        this( 1000 );
    }

    /**
     * Creates a DefaultRemoteIteratorTableModel with a default estimated size of
     * 1000 rows. Anyway, the underlying container will be trimmed to size when
     * all the rows will have been loaded. The provided RemoteIteratorClient is
     * used and the loading is started.
     *
     * @param p_remoteIteratorClient The RemoteIteratorClient to use for data loading.
     *
     * @see #DefaultRemoteIteratorTableModel( int p_estimatedSize )
     * @see RemoteIteratorClient
     */
    public DefaultRemoteIteratorTableModel( RemoteIteratorClient p_remoteIteratorClient ) {
        this();

        // We set theRemoteIteratorClient
        setRemoteIteratorClient( p_remoteIteratorClient );

        // We start to load the data
        startLoadingData();
    }

    /**
     * Creates a DefaultRemoteIteratorTableModel with the given int as the size
     * of the underlying List storing the rows. The increase value will use the
     * given int as well. Anyway, the underlying container will be trimmed to
     * size when all the rows will have been loaded.
     *
     * @param p_estimatedSize The size and the increase number to use for the
     * underlying List storing the* rows. Should be used to prevent on-going
     * resize of the List. No check is done on this value which, obviously,
     * should be greater than 0.
     */
    public DefaultRemoteIteratorTableModel( int p_estimatedSize ) {
        super();

        _rows = new Vector( p_estimatedSize, p_estimatedSize );
    }

    /**
     * Creates a DefaultRemoteIteratorTableModel with the given estimated size of
     * rows. Anyway, the underlying container will be trimmed to size when all
     * the rows will have been loaded. The provided RemoteIteratorClient is used
     * and the loading is started.
     *
     * @param p_remoteIteratorClient The RemoteIteratorClient to use for data loading.
     * @param p_estimatedSize The size and the increase number to use for the
     * underlying List storing the* rows. Should be used to prevent on-going
     * resize of the List. No check is done on this value which, obviously,
     * should be greater than 0.
     *
     * @see #DefaultRemoteIteratorTableModel( int p_estimatedSize )
     * @see RemoteIteratorClient
     */
    public DefaultRemoteIteratorTableModel( RemoteIteratorClient p_remoteIteratorClient, int p_estimatedSize ) {
        this( p_estimatedSize );

        // We set theRemoteIteratorClient
        setRemoteIteratorClient( p_remoteIteratorClient );

        // We start to load the data
        startLoadingData();
    }

    // *************************************************************************
    // ********************** AbstractTableModel Methods ***********************
    // *************************************************************************

    public int getRowCount() {
        return _rows.size();
    }

    public int getColumnCount() {
        if ( getRowCount() > 0 ) {
            // We return the number of column in the first Row
            List aRow = (List) _rows.get( 0 );
            return aRow.size();
        } else {
            return 0;
        }
    }

    public Object getValueAt( int p_rowIndex, int p_columnIndex ) {
        List aRow = (List) _rows.get( p_rowIndex );

        return aRow.get( p_columnIndex );
    }

    // *************************************************************************
    // ****************** RemoteIteratorTableModel Methods *********************
    // *************************************************************************

    public void setRemoteIteratorClient( RemoteIteratorClient p_remoteIteratorClient ) {
        if ( p_remoteIteratorClient == null ) {
            throw new IllegalArgumentException( "Need a non null RemoteIteratorClient..." );
        }

        // If needed, we remove the old rows and notify the listeners
        if ( isLoadingData() ) {
            stopLoadingData();
        }

        // We clear the data already present
        if ( _rows.size() != 0 ) {
            _rows.clear();
            fireTableDataChanged();
        }

        // We instantiate the loading Thread
        _loadingThread = new BackgroundLoadingThread( p_remoteIteratorClient, this );
    }

    public void startLoadingData() {
        if ( _loadingThread != null ) {
            _loadingThread.start();
        } else {
            throw new IllegalStateException( "The data have already been loaded..." );
        }
    }

    public void stopLoadingData() {
        if ( _loadingThread != null ) {
            _loadingThread.stopLoading(); // We request the stop loading...
            _loadingThread = null; // We discard the ref...

            // No more data will come...
            _rows.trimToSize();
        }
    }

    public boolean isLoadingData() {
        boolean isLoadingData = false;

        if ( _loadingThread != null && _loadingThread.isAlive() ) {
            isLoadingData = true;
        }

        return isLoadingData;
    }

    public void releaseResources() {
        if ( _loadingThread != null ) {
            _loadingThread.stopLoading();
            _loadingThread = null;
        }
    }

    // *************************************************************************
    // ****************************** Accessors ********************************
    // *************************************************************************

    /**
     * @return the default block size (the size of the returned block, but the first
     * one).
     *
     * @see #setDefaultBlockSize
     * @see #setInitialBlockSize
     */
    public int getDefaultBlockSize() {
        return _defaultBlockSize;
    }

    /**
     * @see #getDefaultBlockSize
     */
    public void setDefaultBlockSize( int p_defaultBlockSize ) {
        _defaultBlockSize = p_defaultBlockSize;
    }

    /**
     * @return the initialBlockSize (the size of the first returned block, should
     * be small to accelerate data display to the user...).
     *
     * @see #setInitialBlockSize
     */
    public int getInitialBlockSize() {
        return _initialBlockSize;
    }

    /***
     * @see #getInitialBlockSize
     */
    public void setInitialBlockSize( int p_initialBlockSize ) {
        _initialBlockSize = p_initialBlockSize;
    }

    /**
     * @return The delay, in ms, between two calls to next() on the RemoteIterator.
     * Default value is 0.
     *
     * @see #setDelayBetweenRetrieval
     */
    public int getDelayBetweenRetrieval() {
        return _delayBetweenRetrieval;
    }

    /**
     * Sets the delay between each call to the RI (could be used when multiple
     * models are accessing various RI to prevent locks because some model are
     * calling next() all the time...).
     *
     * @param p_delayBetweenRetrieval in ms.
     */
    public void setDelayBetweenRetrieval( int p_delayBetweenRetrieval ) {
        _delayBetweenRetrieval = p_delayBetweenRetrieval;
    }

    // *************************************************************************
    // ***************************** Action Methods ****************************
    // *************************************************************************

    /**
     *
     * @param p_rowIterator should not be null... Used by the background loading
     * thread to add rows to the model.
     */
    public synchronized void addRows( Iterator p_rowIterator ) {
        int firstInsertedRow = getRowCount();

        while ( p_rowIterator.hasNext() ) {
            _rows.add( p_rowIterator.next() );
        }

        // We notify the listeners...
        fireTableRowsInserted( firstInsertedRow, getRowCount() );
    }

    /**
     * Removes all the remaining RI. Usually called when the application is exiting
     * because, even though a stop is requested, some threads might be sleeping and
     * the application will quit BEFORE the RI are removed (it might give some
     * troubles to the server).
     */
    public static void cleanUp() {
        BackgroundLoadingThread.removeAllRemainingRemoteIterator();
    }
}
