/* ******************************************************************************** *
 * Copyright (c) 2002 - 2004 Bright Side Factory. All rights reserved.              *
 *                                                                                  *
 * Redistribution and use in source and binary forms, with or without modification, *
 * are permitted provided that the following conditions are met:                    *
 *                                                                                  *
 * 1. Redistributions of source code must retain the above copyright notice, this   *
 * list of conditions and the following disclaimer.                                 *
 *                                                                                  *
 * 2. Redistributions in binary form must reproduce the above copyright notice,     *
 * this list of conditions and the following disclaimer in the documentation and/or *
 * other materials provided with the distribution.                                  *
 *                                                                                  *
 * 3. The end-user documentation included with the redistribution, if any, must     *
 * include the following acknowledgment: "This product includes software developed  *
 * by the Bright Side Factory (http://www.bs-factory.org/)." Alternately, this      *
 * acknowledgment may appear in the software itself, if and wherever such           *
 * third-party acknowledgments normally appear.                                     *
 *                                                                                  *
 * 4. The names "Bright Side", "BS Framework" and "Bright Side Factory" must not be *
 * used to endorse or promote products derived from this software without prior     *
 * written permission. For written permission, please contact info@bs-factory.org.  *
 *                                                                                  *
 * 5. Products derived from this software may not be called "Bright Side", nor may  *
 * "Bright Side" appear in their name, without prior written permission of the      *
 * Bright Side Factory.                                                             *
 *                                                                                  *
 * THIS SOFTWARE IS PROVIDED ''AS IS'' BY THE COPYRIGHT OWNER AND ANY EXPRESSED OR  *
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF     *
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT *
 * SHALL THE COPYRIGHT OWNER OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,          *
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,   *
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,    *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF  *
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE  *
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS OFTWARE, EVEN IF ADVISED *
 * OF THE POSSIBILITY OF SUCH DAMAGE.                                               *
 *                                                                                  *
 * ================================================================================ *
 *                                                                                  *
 * This software consists of voluntary contributions made by many individuals on    *
 * behalf of the Bright Side Factory. For more information on the Bright Side       *
 * Factory, please see http://www.bs-factory.org.                                   *
 *                                                                                  *
 * ******************************************************************************** */
package org.bsf.remoteIterator.server;

import org.bsf.commons.ejb.SessionAdapterBean;
import org.bsf.remoteIterator.common.ColumnMetadata;
import org.bsf.remoteIterator.common.RemoteIteratorResultTable;

import javax.ejb.CreateException;
import java.sql.*;
import java.util.ArrayList;
import java.util.List;

/**
 * Executes and provides an iterator on a SQL statement using the Bean container
 * to handle passivation and so on... Usually used to populate a Table on the
 * client without transfering all the rows at once.
 *
 * @ejb:bean            type="Stateful"
 *                      name="RemoteIterator"
 *                      jndi-name="ejb/RemoteIterator"
 *                      generate="true"
 *                      view-type="remote"
 *
 * @ejb:home            extends="javax.ejb.EJBHome"
 *
 * @ejb:interface       extends="javax.ejb.EJBObject"
 *
 * @ejb:transaction     type="Supports"
 *
 * @ejb:resource-ref    res-name="jdbc/DBPool"
 *                      res-type="javax.sql.DataSource"
 *                      res-auth="Container"
 *
 *  @jboss:resource-ref res-ref-name="jdbc/DBPool"
 *                      resource-name="FwkDataSourceManager"
 *
 *  @jonas.bean         ejb-name="RemoteIterator"
 *                      jndi-name="ejb/RemoteIterator"
 *
 *  @jonas.resource     res-ref-name="jdbc/DBPool"
 *                      jndi-name="jdbc_1"
 */
public class RemoteIteratorBean extends SessionAdapterBean {
    // *************************************************************************
    // *************************** Static variables ****************************
    // *************************************************************************

    // Various constants (self explanatory)
    private static final int RI_BUFFER_SIZE = 2000;
    private static final int RI_DEFAULT_BLOCK_SIZE = 150;

    private static final String EMPTY_STRING = "";
    private static final String JNDI_DATASOURCE_NAME = "DBPool";

    // *************************************************************************
    // *************************** Instance variables **************************
    // *************************************************************************

    /**
     * Keeps the size of the last read block.
     */
    private int _lastReadBlockSize = 0;

    /**
     * The curent position into the resultset (needed when passivated or
     * activated).
     */
    private int _position = 0;

    private String _SQLQuery;
    private List _columnMetadatas;
    private int _defaultBlockSize = RI_DEFAULT_BLOCK_SIZE;

    // TRANSIENT FIELDS (Cnx informations, we need them... :o)
    private transient Statement _statement;
    private transient ResultSet _resultSet;
    private transient Connection _connection;

    // *************************************************************************
    // *************************** Utility Methods *****************************
    // *************************************************************************

    /**
     * Creates the meta data from the actual resultsets.
     */
    private void createMetadatas() {
        _columnMetadatas = new ArrayList();

        try {
            ResultSetMetaData resultSetMetaData = _resultSet.getMetaData();

            int nbColums = resultSetMetaData.getColumnCount();

            for ( int col = 1 ; col <= nbColums ; col++ ) {
                ColumnMetadata temp = new ColumnMetadata();

                temp.setColumnClassName( resultSetMetaData.getColumnClassName( col ) );
                temp.setColumnName( resultSetMetaData.getColumnName( col ) );
                temp.setSchemaName( resultSetMetaData.getSchemaName( col ) );
                temp.setTableName( resultSetMetaData.getTableName( col ) );
                temp.setColumnType( resultSetMetaData.getColumnType( col ) );
                temp.setColumnTypeName( resultSetMetaData.getColumnTypeName( col ) );
                temp.setPrecision( resultSetMetaData.getPrecision( col ) );
                temp.setScale( resultSetMetaData.getScale( col ) );

                _columnMetadatas.add( temp );
            }
        } catch( SQLException exception ) {
            handleSQLException( exception );
        }
    }

    /**
     * @param p_query The SQL statement to execute. It should be a valid SQL statement
     * otherwise you'll get an SQL exception later :o) !!!
     *
     * @throws java.lang.RuntimeException if the given query is null or blank.
     *
     * @ejb:create-method
     */
    public void ejbCreate( String p_query ) throws CreateException {
        logGraphBegin( "create" );

        // We check if the given p_query is valid...
        if ( p_query == null || EMPTY_STRING.equals( p_query ) ) {
            handleExceptionAsSystemException( new RuntimeException( "The p_query shouldn't be null !!!" ) );
        }

        // We whall keep the query
        _SQLQuery = p_query;

        // Let's go for a first execution of the SQL statement
        firstQueryExecution();

        // We create the MetaData (could be requested by the client)
        createMetadatas();

        logGraphEnd( "create" );
    }

    /**
     * Ensures that everything is closed correctly (mainly the DB ressources).
     */
    public void ejbRemove() {
        logGraphBegin( "ejbRemove " );

        // close connection
        closeDataSource();

        logGraphEnd( "ejbRemove" );
    }

    // *************************************************************************
    // ******************************* Accessors *******************************
    // *************************************************************************

    /**
     * @return A List of ColumnMetaData corresponding to the executed SQL statement.
     *
     * @see org.bsf.remoteIterator.common.ColumnMetadata
     *
     * @ejb:interface-method
     */
    public List getColumnMetaData() {
        return _columnMetadatas;
    }

    /**
     * @return the block of rows starting at the given position and of the default
     * block size.
     *
     * @see #absolute(int p_position, int p_blockSize)
     * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
     *
     * @ejb:interface-method
     */
    public RemoteIteratorResultTable absolute( int p_position ) {
        return absolute( p_position, RI_DEFAULT_BLOCK_SIZE );
    }

    /**
     * @return the block of rows starting at the given position and of the given
     * size.
     *
     * @see #absolute(int p_position)
     * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
     *
     * @ejb:interface-method
     */
    public RemoteIteratorResultTable absolute( int p_position, int p_blockSize ) {
        try {
            _resultSet.absolute( p_position );
        } catch( SQLException sqlException ) {
            // Unable to reach the desired position...
            handleSQLException( sqlException );
        }

        // We managed to set the cursor at the desired position, let's return the rows...
        return next( p_blockSize );
    }

    /**
     * @return the previous "block" of rows (starting from before the first row
     * of the actual block) as a RemoteIteratorResultTable. The default block
     * size is used. Returns the default block size rows from the begining
     * if the default block size to use is bigger than the available rows.
     *
     * @see #previous(int p_blockSize)
     * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
     *
     * @ejb:interface-method
     */
    public RemoteIteratorResultTable previous() {
        return previous( _defaultBlockSize );
    }

    /**
     * @param p_blockSize the size of the block to return. Should be greater than
     * zero (not checked).
     *
     * @return the previous p_blockSize rows (starting from the actual position) as a
     * RemoteIteratorResultTable. Return to the beginning of the actual block before
     * returning the previous rows. Returns the p_blockSize rows from the begining
     * if the given block size is bigger than the available rows.
     *
     * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
     *
     * @ejb:interface-method
     */
    public RemoteIteratorResultTable previous( int p_blockSize ) {
        String logMessage = "previous(" + p_blockSize + ")";
        logGraphBegin( logMessage );

        // Creating a RemoteIteratorResultTable
        RemoteIteratorResultTable remoteIteratorResultTable = new RemoteIteratorResultTable( 0 );

        try {
            // We have to move to (begining of the actual block - block size)
            int positionToReach = _resultSet.getRow() - _lastReadBlockSize - p_blockSize;

            if ( positionToReach <= 0 ) {
                _resultSet.beforeFirst();
            } else {
                _resultSet.absolute( positionToReach );
            }
            remoteIteratorResultTable = next( p_blockSize );
        } catch( SQLException e ) {
            // That might happen...
            handleSQLException( e );
        }

        logGraphEnd( logMessage );

        return remoteIteratorResultTable;
    }

    /**
     * @return the next "block" of rows (starting from the actual position) as a
     * RemoteIteratorResultTable. The default block size is used.
     *
     * @see #next(int p_blockSize)
     * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
     *
     * @ejb:interface-method
     */
    public RemoteIteratorResultTable next() {
        return next( _defaultBlockSize );
    }

    /**
     * @param p_blockSize the size of the block to return. Should be greater than
     * zero (not checked).
     *
     * @return the next p_blockSize rows (starting from the actual position) as a
     * RemoteIteratorResultTable.
     *
     * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
     *
     * @ejb:interface-method
     */
    public RemoteIteratorResultTable next( int p_blockSize ) {
        String logMessage = "next(" + p_blockSize + ")";
        logGraphBegin( logMessage );

        int readBlockPosition = 0;
        boolean willBeFirst = false;

        // Creating a RemoteIteratorResultTable
        RemoteIteratorResultTable remoteIteratorResultTable = new RemoteIteratorResultTable( p_blockSize );

        try {
            if ( isBeforeFirst() ) {
                // We know that after the move we will be at the end of the first block
                willBeFirst = true;
            }

            if ( isAfterLast() ) {
                // We have already reached the end of data
                remoteIteratorResultTable.setAlreadyLast( true );
                remoteIteratorResultTable.setLast( true );

                // No need to go on
                return remoteIteratorResultTable;
            }

            // We can read the rows requested rows

            // The lastNotReached test is needed because the isAfterLast returns
            // false if there is no data...
            boolean lastNotReached = true;

            while ( ( readBlockPosition < p_blockSize ) && ( lastNotReached = _resultSet.next() ) ) {
                // Found one... Add it to the RemoteIteratorResultTable...
                addCurrentResultSetRow( remoteIteratorResultTable );

                // We increment the position in the current block (ie. Next please !!!)
                readBlockPosition++;

                // if last record was reached, we set thtotal size
                if ( isLast() ) {
                    logDebug( "end position found :" + getPosition() );
                }
            }

            // We keep the size of the last read block
            _lastReadBlockSize = readBlockPosition;

            // If needed, we notify that we have read the first record (ie. the block is containing the first record)
            if ( willBeFirst ) {
                remoteIteratorResultTable.setFirst( ( true ) );
            }

            //  If needed, we notify that we have read the last record
            if ( !lastNotReached || isAfterLast() || isLast() ) {
                remoteIteratorResultTable.setLast( ( true ) );
            }
        } catch( SQLException e ) {
            // That might happen on the next...
            handleSQLException( e );
        }

        logGraphEnd( logMessage );

        return remoteIteratorResultTable;
    }

    /**
     * @return the total number of records (the cursor remains on its current
     * position).
     *
     * @ejb:interface-method
     */
    public Long getRowCount() {
        logGraphBegin( "getRowCount" );

        long totalSize = 0;

        try {
            // We retrieve the current position (to restore it, cf. below)
            int currentPosition = getPosition();

            // We move to the last record
            _resultSet.last();

            // We get the last record position;
            totalSize = getPosition();

            // We restore the cursor previous position
            if ( currentPosition == 0 ) {
                _resultSet.beforeFirst();
            } else {
                _resultSet.absolute( currentPosition );
            }
        } catch( SQLException eSQL ) {
            handleSQLException( eSQL );
        }

        logGraphEnd( "getRowCount: found " + totalSize );

        return new Long( totalSize );
    }

    /**
     * @return true if the current block is the last block.
     */
    private boolean isLast() throws SQLException {
        return _resultSet.isLast();
    }

    /**
     * @return true if we are after the last record
     */
    private boolean isAfterLast() throws SQLException {
        return _resultSet.isAfterLast();
    }

    /**
     * @return true if the cursor is before the first record, false otherwise.
     */
    private boolean isBeforeFirst() throws SQLException {
        return _resultSet.isBeforeFirst();
    }

    /**
     * @return The current position of the cursor in the ResultSet, 0 if the
     * resultset is null.
     */
    private int getPosition() throws SQLException {
        if ( _resultSet == null )
            return 0;
        else
            return _resultSet.getRow();
    }

    /**
     * Sets the cursor to the specified position and update the lastReadBlockSize.
     */
    private void restorePosition( int p_position, int p_lastReadBlockSize ) throws SQLException {
        if ( p_position > 0 ) {
            _resultSet.absolute( p_position );
        }

        _lastReadBlockSize = p_lastReadBlockSize;
    }

    /**
     * Adds the current record to the given RemoteIteratorResultTable.
     *
     * @throws java.sql.SQLException if there's a problem in the retrieval of the record...
     * @throws java.lang.NullPointerException if the given ResultTable is null.
     *
     * @see org.bsf.remoteIterator.common.RemoteIteratorResultTable
     */
    private void addCurrentResultSetRow( RemoteIteratorResultTable p_resultTable ) throws SQLException {
        int columnCount = _resultSet.getMetaData().getColumnCount();
        ArrayList row = new ArrayList( columnCount );

        for ( int columnIndex = 0 ; columnIndex < columnCount ; columnIndex++ ) {
            // We add the default object returned by the driver
            row.add( _resultSet.getObject( columnIndex + 1 ) );
        }

        // And we add the created row to the ResultTable
        p_resultTable.addRow( row, new Integer( _resultSet.getRow() ) );
    }

    /**
     * Must be called before executeQuery as it creates the statement and configures
     * it.
     */
    private void prepareQuery() throws SQLException {
        logGraphBegin( "prepareQuery" );

        // We create a new statement (scroll insensible)
        _statement = _connection.createStatement( ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY );

        // To optimize the fetch (could be tuned)
        _statement.setFetchSize( RemoteIteratorBean.RI_BUFFER_SIZE );

        logGraphEnd( "prepareQuery" );
    }

    /**
     * Prepares and executes the SQL query.
     *
     * @throws java.sql.SQLException if there's a problem while preparing or executing the SQL.
     */
    private void executeQuery() throws SQLException {
        logGraphBegin( "executeQuery \nQuery : " + _SQLQuery );

        try {
            // We prepare the query (cf. above)
            prepareQuery();

            // We reset the logical position information
            _lastReadBlockSize = 0;

            if ( _statement == null )
                throw new SQLException( "Statement null ! You must call prepareQuery() before any executeQuery() ! : " );

            // We execute query
            _resultSet = _statement.executeQuery( _SQLQuery );
        } catch( SQLException sqlException ) {
            logFatal( "executeQuery : error while executing query " + _SQLQuery, sqlException );

            throw sqlException;
        }
    }

    /**
     * Should be called for the first execution of the Query (opens the DataSource
     * and executes the query).
     */
    private void firstQueryExecution() {
        // Opens the database and executes the query
        try {
            openDataSource();

            executeQuery();
        } catch( SQLException Se ) {
            // We cleanup the situation before throwing the exception
            closeDataSource();
            handleSQLException( Se );
        }
    }

    /**
     * Retrieves a connection to the Database.
     */
    private void openDataSource() throws SQLException {
        logGraphBegin( "openDataSource" );

        _connection = getConnection( JNDI_DATASOURCE_NAME );
    }

    /**
     * Closes the DB ressources (ResultSet, Statement, Connection).
     */
    private void closeDataSource() {
        logGraphBegin( "closeDataSource" );

        try {
            if ( _resultSet != null ) _resultSet.close();
        } catch( SQLException Se ) {
            logError( Se.getLocalizedMessage() );
        }

        try {
            if ( _statement != null ) _statement.close();
        } catch( SQLException e ) {
            logError( e.getLocalizedMessage() );
        }

        try {
            if ( _connection != null ) _connection.close();
        } catch( SQLException e ) {
            logError( e.getLocalizedMessage() );
        }

        logGraphEnd( "closeDataSource" );
    }

    /**
     * Sets the default block size to be used for this RemoteIteratorBean.
     *
     * @ejb:interface-method
     */
    public void setDefaultBlockSize( Integer p_defaultBlockSize ) {
        if ( p_defaultBlockSize == null )
            _defaultBlockSize = RI_DEFAULT_BLOCK_SIZE;
        else
            _defaultBlockSize = p_defaultBlockSize.intValue();
    }

    /////////////////////////////////////////////////////////////////////////
    // TECHNICAL CONTRACT
    /////////////////////////////////////////////////////////////////////////

    public void ejbActivate() {
        logGraphBegin( "ejbActivate" );

        try {
            // Save postion before being reinitialized by prepareQuery;
            int oldPosition = _position;
            int oldLastReadBlockSize = _lastReadBlockSize;

            // Restoring the connection
            firstQueryExecution();

            // We restore the "old" position
            restorePosition( oldPosition, oldLastReadBlockSize );
        } catch( SQLException Se ) {
            handleSQLException( Se );
        }

        logDebug( " State after reactivation : lastReadBlockSize = " + _lastReadBlockSize + ", position = " + _position );

        logGraphEnd( "ejbActivate " );
    }

    public void ejbPassivate() {
        logGraphBegin( "ejbPassivate " );

        try {
            if ( _resultSet != null ) {
                // We keep the current position (for activation...)
                _position = getPosition();
            } else {
                _position = 0;
            }

            // We release the connection
            closeDataSource();
        } catch( SQLException e ) {
            handleSQLException( e );
        }

        logDebug( " State for passivation: lastReadBlockSize = " + _lastReadBlockSize + ", position = " + _position );
        logGraphEnd( "ejbPassivate" );
    }
}